/*===========================================================================*\
|| Copyright 1988,1989,1990,1991 Pittsburgh Powercomputing Corporation (PPc) 
|| of Pittsburgh, Pennsylvania as an unpublished work. All Rights Reserved.
||
|| This is unpublished proprietary trade secret source code of Pittsburgh
|| Powercomputing. The copyright above does not evidence any actual or intended
|| publication of this source code.  This source code is supplied under a
|| license.  It may be used, disclosed, and/or copied only as permitted under
|| such license agreement.  Use, copying, and/or disclosure of this source
|| code is strictly prohibited unless otherwise provided in the license
|| agreement.  Any copy must contain the copyright above and this notice must
|| be preserved in all copies of this source code.
\*===========================================================================*/
/*
 
 $Header: /usr/ppc/cvsroot/x11r4/server/diag/binpatch.c,v 1.1 91/04/28 19:24:57 ppc Exp $

 $Log:	binpatch.c,v $
 * Revision 1.1  91/04/28  19:24:57  ppc
 * Initial revision
 * 

*/

#ifndef lint
static char cvsid[] = "$Id: binpatch.c,v 1.1 91/04/28 19:24:57 ppc Exp $ Copyright 1991 PPc";
#endif
/* Change bytes in an binary file. Dan Christian 4-10-91 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

static char *progName;
static int quiet=0;

usage()
{
    fprintf (stderr,"Usage: %s [switches] <file> <offset> <byte> [byte...] \n",
	    progName);
    fprintf (stderr,"  The offset or bytes may be hex, octal, or decimal.\n");
    fprintf (stderr,"-quiet:  Don't show data or ask for verification.\n");
    fprintf (stderr," -fast:  Don't ask for verification.\n");
    fprintf (stderr," -help:  Display this message.\n");
    fprintf (stderr,"  -hex:  Assume all arguments are hex numbers.\n");

    exit (2);
}

main (argc, argv)
  int	argc;
  char	**argv;
{
    unsigned char	data[4096], *dp, old[16];
    long	offset, verify;
    int		length, tmp, fd;
    int		fast=0, hex=0;
    char	*file, buff[256];

    progName = argv[0];

    for (++argv, --argc; argc; ++argv, --argc) {
	if (**argv != '-') break;
	if (!strncmp (*argv, "-fast", 2)) ++fast;
	else if (!strncmp (*argv, "-hex", 4)) ++hex;
	else if (!strncmp (*argv, "-quiet", 2)) ++quiet;
	else if (!strncmp (*argv, "-help", 4)) usage ();
	else {
	    fprintf (stderr,
		     "Argument %s ignored.  Use '%s -help' for usage.\n",
		    *argv, progName);
	}
    }
    if (argc < 3) {
	fprintf (stderr,"Additional arguments expected.\n");
	usage ();
    }
    file = *argv++; --argc;
    if (sscanf (*argv, (hex) ? "%lx" : "%li", &offset) < 1) {
	fprintf (stderr,"Error parsing offset <%s>\n", *argv);
	usage ();
    }
    if (offset < 0) {
	fprintf (stderr,"Invalid offset %d\n", offset);
	usage ();
    }
    ++argv; --argc;
				/* get bytes to change */
    length=0;
    dp = data;
    for (; argc; ++argv, --argc, ++dp, ++length) {
	if (sscanf (*argv, (hex) ? "%x" : "%i", &tmp) < 1) {
	    fprintf (stderr,"Error parsing data byte <%s>\n", *argv);
	    usage ();
	}
	if ((tmp > 255) || (tmp < 0)) {
	    fprintf (stderr,"Invalid byte: %d\n", tmp);
	    usage ();
	}
	*dp = tmp;
    }

    if (!quiet)
      printf ("Changing %d bytes at 0x%lx in <%s>\n", length, offset, file);

    if ((fd = open (file, O_RDWR)) < 0) {
	fprintf (stderr, "%s: Could not open file <%s>.\n", progName, file);
	perror ("open");
	exit (1);
    }

    if (!quiet) {
	verify = offset-2;
	if (verify < 0) verify = 0;
	if (lseek (fd, verify, SEEK_SET) == -1) {
	    fprintf (stderr, "%s: Error seeking to 0x%x\n", progName, verify);
	    perror ("lseek");
	    exit (1);
	}
	read (fd, old, 8);
	printf ("Old values are:\n");
	printf ("0x%x: \t%02x %02x  ** %02x **  %02x %02x %02x %02x %02x\n",
		verify, old[0], old[1], old[2], old[3],
		old[4], old[5], old[6], old[7]);
	
	if (!fast) {		/* Get confirmation first */
	    printf ("Do you want to change this value to 0x%02x [y]? ",
		    data[0]);
	    fflush (stdout);
	    fgets (buff, sizeof (buff), stdin);
	    if ((buff[0] != 'y') && (buff[0] != 'Y') &&
		(buff[0] != '\n') && (buff[0] != '\r')) {
		printf ("Aborting binary patch.\n");
		close (fd);
		exit (0);
	    }
	}
    }
    /* Write the data */
    if (lseek (fd, offset, SEEK_SET) == -1) {
	fprintf (stderr, "%s: Error seeking to 0x%x\n", progName, offset);
	perror ("lseek");
	exit (1);
    }
    write (fd, data, length);
    
    if (!quiet) {
	/* verify for user */
	if (lseek (fd, verify, SEEK_SET) == -1) {
	    fprintf (stderr, "%s: Error seeking to 0x%x\n", progName, verify);
	    perror ("lseek");
	    exit (1);
	}
	read (fd, old, 8);
	printf ("New values are:\n");
	printf ("0x%x: \t%02x %02x  ** %02x **  %02x %02x %02x %02x %02x\n",
		verify, old[0], old[1], old[2], old[3],
		old[4], old[5], old[6], old[7]);
    }
 
    close (fd);
    exit (0);
}
