/* Dan Christian */

#include <X/Xlib.h>
#include <libezX.h>
#include <stdio.h>
#include <math.h>

#define  PI 3.141592654

void drawAxis();
void openBasic ();
void print_debug ();
void plotRaw ();

int DEBUG = 0;			/* global debugging level flag */

/* draw an axis in the current window. the axis spans the window.
 * axis is a character for which axis to draw ('x' or 'y')
 * center is where the 0 should be in the window (in pixels)
 * step is the increment for tick marks (in pixels, the decimal portion
 *        is truncated)
 * lstep is the increment for labeling tick marks, 0 if no labeling is used.
 * field is the field width for labeling.  If lstep == 0, this is not used.
 */
void drawAxis (axis, cx, cy, step, lstep, field)
  char axis;
  int cx, cy, field;
  double step, lstep;
{
    double p, t;
    int tick, width, height, alternate;
    char txt[32];

    if (DEBUG > 2) printf ("Entered drawAxis: %c %d %d %f %f %d\n",
			   axis, cx, cy, step,lstep, field);
    tick = 4;
    height = EZX_GraphicsHeight ();
    width  = EZX_GraphicsWidth ();

    step = fabs (step);

    if (axis == 'x') {
	print_debug (2, "Drawing x axis\n");
	EZX_DrawLine (0, cy, width, cy); /* draw x axis */

	alternate = 1;
	for (p = cx + step, t = lstep; p < width; t += lstep, p += step) {
	    EZX_DrawLine ((int) p, cy, (int) p, cy + tick);
	    if ((lstep !=0) && (alternate)) {
		sprintf (txt, "%*g", field, t);
		EZX_FormatAt ((int) p, cy + tick + 1, txt, 'c');
	    }
	    alternate = alternate ? 0 : 1;
	}
	alternate = 1;
	for (p = cx - step, t = -lstep; p > 0; t -= lstep, p -= step) {
	    EZX_DrawLine ((int) p, cy, (int) p, cy + tick);
	    if ((lstep !=0) && (alternate)) {
		sprintf (txt, "%*g", field, t);
		EZX_FormatAt ((int) p, cy + tick + 1, txt, 'c');
	    }
	    alternate = alternate ? 0 : 1;
	}
    } else {
	print_debug (2, "Drawing y axis\n");
	EZX_DrawLine (cx, 0, cx, height); /* draw y axis */

	alternate = 1;
	for (p = cy + step, t = -lstep; p < height; t -= lstep, p += step) {
	    EZX_DrawLine (cx, (int) p, cx - tick, (int) p);
	    if ((lstep !=0) && (alternate)) {
		sprintf (txt, "%*g", field, t);
		EZX_FormatAt (cx - tick - 1, (int) p, txt, 'r');
	    }
	    alternate = alternate ? 0 : 1;
	}
	alternate = 1;
	for (p = cy - step, t = lstep; p > 0; t += lstep, p -= step) {
	    EZX_DrawLine (cx, (int) p, cx - tick, (int) p);
	    if ((lstep !=0) && (alternate)) {
		sprintf (txt, "%*g", field, t);
		EZX_FormatAt (cx - tick - 1, (int) p, txt, 'r');
	    }
	    alternate = alternate ? 0 : 1;
	}
    }
}

void
makeWindow(sizex, sizey, cpx, cpy, tx, ty, scalex, scaley)
  int sizex, sizey, cpx, cpy, tx, ty;
  double scalex, scaley;
{
    EZX_Window_t *window;	/* pointer to the window */
    int cx, cy;
    int h, w;
    float   temp, x, step;
    
    print_debug (3, "entered OpenBasic\n");
 
    window = EZX_MakeWindow (sizex, sizey, "ezPlot Window");
    EZX_UseWindow (window);
    
    h = EZX_GraphicsHeight ();
    w = EZX_GraphicsWidth ();
    cx = w * cpx / 100.0;
    cy = h - h * cpy / 100.0;
    
    drawAxis ( 'x', cx, cy, (float)w/tx, (double)scalex/tx, 4);
    drawAxis ( 'y', cx, cy, (float)h/ty, (double)scaley/ty, 4);
    
    EZX_BatchOff();
    if (DEBUG == 0) EZX_BatchOn ();
    print_debug (2, "Created window\n");
}

void
plotRaw (input, cpx, cpy, scalex, scaley)
  FILE *input;
  float scalex, scaley;
  int cpx, cpy;
{
    float x, y, last_x, last_y, sx, sy;
    int height, width, cx, cy;

    print_debug (3, "Entered plotRaw\n");
       
    height = EZX_GraphicsHeight ();
    width  = EZX_GraphicsWidth ();
    cx = width * cpx / 100.0;
    cy = height - height * cpy / 100.0;
    sx = width / scalex;
    sy = height / scaley;

    if ( fscanf (input, "%f %f", &x, &y) == 2) {
	 last_x = cx + x * sx;
	 last_y = cy - y * sy;
	 while ( fscanf (input, "%f %f", &x, &y) == 2) {
	     EZX_DrawLine ((int)last_x, (int)last_y,
			   (int)(cx + x * sx), (int)(cy - y * sy));
	     last_x = cx + x * sx;
	     last_y = cy - y * sy;
	 }
     }
    fclose (input);
}

void print_debug (level, string)
  int level;
  char *string;
{
    if (DEBUG > level)
      fprintf (stderr, string);
}

main (argc, argv)
int argc;
char **argv;
{
    int i;
    float scalex, scaley;
    int centerx, centery;
    FILE *input;
    char *infile;		/* an input file name */

    scalex = 100;		/* default scale */
    scaley = 100;
    centerx = centery = 50;
    infile = NULL;

    for (i = 1; i < argc; i++)	/* parse command line for flags */
      if (argv[i][0] == '-')
	switch (argv[i][1]) {
	    case 'X':		/* debuging level 0-9*/
	    DEBUG = argv[i][2] - '0';
	    break;

	    case 's':
	    if (argv[i][2] == 'x')
	      sscanf (argv[++i], "%f", &scalex);
	    else if (argv[i][2] == 'y')
	      sscanf (argv[++i], "%f", &scaley);
	    break;

	    case 'c':
	    if (argv[i][2] == 'x')
	      sscanf (argv[++i], "%d", &centerx);
	    else if (argv[i][2] == 'y')
	      sscanf (argv[++i], "%d", &centery);
	    break;

	    case 'h':
	    printf ("Valid arguments: -Xn, -sx -sy -cx -cy.\n");
	    exit (0);

	    default:
	    printf
	      ("Valid arguments: -Xn, -sx -sy -cx -cy.  No: %s.\n", argv[i]);
	    exit (1);
	}
    else
      infile = argv[i];

    if (infile == NULL)
	input = stdin;
    else {
	input = fopen (infile, "r");
	if (input == NULL) {
	    fprintf (stderr, "Unable to open file %s.\n", infile);
	    exit (1);
	}
    }

    EZX_InitGraphics (NULL, argv[0]); /* Initialize and set window */

    makeWindow (400, 400, centerx, centery, 20, 20, scalex, scaley);
    plotRaw (input, centerx, centery, scalex, scaley);

    EZX_BatchOff ();		/* flush buffer */
    printf ("Waiting for mouse click to end\n");
    EZX_WaitForMouseClick ();

    EZX_EndGraphics();
}

