/* Dan Christian, 11-10-1991 */
/*
 
  $Header:$
  $Id:$
  $Log:$
  7-12-92 Dan Christian
  Added built in copy and support for symbolic links.
  Put match/nomatch tests firsts to avoid extra stat calls.
  Added max depth argument
*/
/* traverse two trees, comparing and updating files */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>			/* for fork...  */
#include <sys/wait.h>			/* for waitpid */
#include <utime.h>			/* for utime */
#include <string.h>			/* for strncmp */
#include <limits.h>			/* for PATH_MAX */

/*#ifndef sun
#define lstat stat
#endif*/
/*#define bcopy(to,from,n) memcpy(to,from,n)*/

#define MAXARG	64			/* Number of exec command line args */
#define MAXFILE	256			/* Maximum path length */
char *matchExpr[MAXARG];		/* match expression string list */
char *noMatchExpr[MAXARG];		/* no-match expression string list */
char fastCopy[]="fastCopy";		/* special key argument */
char firstFile[]="first";		/* special key argument */
char secondFile[]="second";		/* special key argument */
char thirdFile[]="third";		/* special key argument */
char *fileExt = NULL, *fileRext;
char *newCmd[MAXARG], *updateCmd[MAXARG], *newdCmd[MAXARG], *allCmd[MAXARG];
int verbose = 0, noExec=0;
int doOwners = 0;			/* set uid/gid */
int noSym = 0;				/* ignore symbolic links */
int symFile = 0;			/* dereference symbolic link files */
int symDir = 0;				/* dereference sysbolic link dirs */
int lnMode = 0;				/* make links instead of copies */
extern int errno;
/*extern	char *sprintf(char *, const char *, ...);*/
int
ParseCmd ( int	argc,
	   char	**argv,
	   char	**argList);
int
MapTree (char	*dir,
	 char	*dir2,
	 int	maxDepth);
int
SubMatch ( char	*pattern,
	   char *string);
int
Copy ( char *fromFile, char *toFile,
    int	newFile,			/* 1 if toFile does not exits */
    struct stat fromStat);
int
RunIt (char *file1, char *file2, char **cline, char *file3);
int
Link ( char *fromFile, char *toFile);

void
usage () {
    fprintf (stderr,
"Usage: maptree tree1 tree2 -update CMD -newd CMD [switches...]\n\
    -update CMD:  Command to run if a tree1 file is newer than a tree2 file.\n\
                  By default, make an identical copy.\n\
       -new CMD:  Command to run if a tree2 file does not exist.\n\
                  By default, the same as update.\n\
      -newd CMD:  Command to run if a tree2 directory does not exist.\n\
                  By default, mkdir {}2.\n\
       -all CMD:  Command to run on all files.\n\
           -uid:  Try to match UID and GID on new files.\n\
           -not:  Show commands but do not execute them.\n\
            -ln:  Make links of new files instead of copies. (no updates)\n\
            -ln:  (used twice) Make links of new directories too.\n\
         -nodef:  Disable default commands for -update and -newd.\n\
         -nonew:  Do nothing if no -new command is specified.\n\
         -nosym:  Ignore all symbolic links.\n\
       -symfile:  Follow symbolic links to files.\n\
        -symdir:  Follow symbolic links to directories.\n\
       -verbose:  Show actions (-v).\n\
   -max <depth>:  Don't go more than depth directories deep.\n\
     -ext <ext>:  Add <Ext> to second file name checks.\n\
    -rext <ext>:  Remove <Ext> from second file checks.\n\
  -match <ptrn>:  Only check files matching Pattern (only * and ? are wild).\n\
-nomatch <ptrn>:  Do not check files matching Pattern.\n\n\
    CMD is a command to run ending with a \\;.\n\
\tThe argument {}1 represents the first file name.\n\
\tThe argument {}2 represents the base second file name.\n\
\tThe argument {}3 represents the second file name as matched and\n\
\t  the ext or rext modifiers have been applied.\n\
\tThe argument {} represents the first file name.\n\
\tAddition occurences of {} represent the second file name.\n\
\tQuirk: csh interprets {}1 entirely differently from {}.\n\
The -match -nomatch and -verbose switches may be given multiple times.\n");
    exit (2);
}

int
main (argc, argv)
  int	argc;
  char	**argv;
{
    char	*first = NULL, *second = NULL;
    char	**cpp, **matchPtr=matchExpr, **noMatchPtr=noMatchExpr;
    int		ret, i;
    int		noDef=0, noNew=0, sawNew=0, sawNewd=0, sawUpdate=0, sawAll=0;
    int		maxDepth=255;

    if (argc < 3) usage ();
    newCmd[0] = NULL;
    updateCmd[0] = NULL;
    newdCmd[0] = NULL;
    allCmd[0] = NULL;
    
    for (--argc, ++argv; argc > 0; ) {
	if (**argv != '-') {
	    if (!first) {
		first = *argv;
	    } else if (!second) {
		second = *argv;
	    } else {
		fprintf (stderr, "maptree: Extraneous tree: %s\n", *argv);
		usage ();
	    }
	    ++argv, --argc;
	} else if (!strncmp (*argv, "-all", 2)) { /* command for new files */
	    ++sawAll; --argc; ++argv;
	    if ((ret = ParseCmd (argc, argv, allCmd)) > 2) {
		argc -= ret; argv += ret;
	    } else {
		fprintf (stderr, "maptree: Error parsing -all command.\n");
		usage ();
	    }
	} else if (!strcmp (*argv, "-new")) { /* command for new files */
	    ++sawNew; --argc; ++argv;
	    if ((ret = ParseCmd (argc, argv, newCmd)) > 2) {
		argc -= ret; argv += ret;
	    } else {
		fprintf (stderr, "maptree: Error parsing -new command.\n");
		usage ();
	    }
	} else if (!strcmp (*argv, "-newd")) { /* command for new dirs */
	    ++sawNewd; --argc; ++argv;
	    if ((ret = ParseCmd (argc, argv, newdCmd)) > 2) {
		argc -= ret; argv += ret;
	    } else {
		fprintf (stderr, "maptree: Error parsing -newd command.\n");
		usage ();
	    }
	} else if (!strncmp (*argv, "-update", 3)) { /* command for update */
	    ++sawUpdate; --argc; ++argv;
	    if ((ret = ParseCmd (argc, argv, updateCmd)) > 2) {
		argc -= ret; argv += ret;
	    } else {
		fprintf (stderr, "maptree: Error parsing -update command.\n");
		usage ();
	    }
	} else if (!strncmp (*argv, "-ln", 3)) { /* make links */
	    ++lnMode; ++argv; --argc;
	} else if (!strncmp (*argv, "-symfile", 5)) { /* sym links as files */
	    ++symFile; ++argv; --argc;
	} else if (!strncmp (*argv, "-symdir", 5)) { /* sym links as files */
	    ++symDir; ++argv; --argc;
	} else if (!strncmp (*argv, "-nosym", 4)) { /* ignore sym links */
	    ++noSym; ++argv; --argc;
	} else if (!strncmp (*argv, "-verbose", 2)) { /* verbosity */
	    ++verbose; ++argv; --argc;
	} else if (!strncmp (*argv, "-uid", 4)) { /* Set UID */
	    ++doOwners; ++argv; --argc;
	} else if (!strncmp (*argv, "-not", 4)) { /* no execute */
	    ++noExec; ++argv; --argc;
	} else if (!strncmp (*argv, "-nodef", 4)) { /* no default actions */
	    ++noDef; ++argv; --argc;
	} else if (!strncmp (*argv, "-nonew", 4)) { /* no default new */
	    ++noNew; ++argv; --argc;
	} else if (!strncmp (*argv, "-max", 4)) { /* max depth */
	    ++argv; --argc;
	    if (!argc || (sscanf (*argv, " %i ", &maxDepth) != 1)) {
		fprintf (stderr, "maptree: Maximum depth expected.\n");
		usage ();
	    }
	    ++argv; --argc;
	} else if (!strncmp (*argv, "-ext", 2)) { /* extension */
	    ++argv; --argc;
	    if (!argc) {
		fprintf (stderr, "maptree: Extension argument expected.\n");
		usage ();
	    }
	    fileExt = *argv;
	    ++argv; --argc;
	} else if (!strncmp (*argv, "-rext", 3)) { /* extension */
	    ++argv; --argc;
	    if (!argc) {
		fprintf (stderr, "maptree: Extension argument expected.\n");
		usage ();
	    }
	    fileRext = *argv;
	    ++argv; --argc;
	} else if (!strncmp (*argv, "-match", 4)) { /* match pattern */
	    ++argv; --argc;
	    if (!argc) {
		fprintf (stderr, "maptree: Match pattern expected.\n");
		usage ();
	    }
	    *matchPtr++ = *argv;
	    ++argv; --argc;
	} else if (!strncmp (*argv, "-nomatch", 4)) { /* NO-match pattern */
	    ++argv; --argc;
	    if (!argc) {
		fprintf (stderr, "maptree: No-Match pattern expected.\n");
		usage ();
	    }
	    *noMatchPtr++ = *argv;
	    ++argv; --argc;
	} else {
	    fprintf (stderr, "maptree: Unknown argument: %s\n", *argv);
	    usage ();
	}
    }
    if (!first || !second) {
	fprintf (stderr, "maptree: Two directories must be specified\n");
	usage ();
    }
    if (lnMode) {			/* check directory formats */
	int lnOK = 0;
	if (*first == '/') ++lnOK;	/* absolution source directory */
	else {				/* see if they have a common root */
	    char *fp, *tp;
	    for (fp = first, tp = second; *fp && (*fp == *tp); ++fp, ++tp)
		if (*fp == '/') {
		    ++lnOK;
		    break;
		}
	}
	if (!lnOK) {
	    fprintf (stderr,
		     "maptree:  invalid source directory(s) for -ln mode <%s> <%s>\n",
		     first, second);
	    fprintf (stderr,
		     "Must either be absolute (/foo) or have a common root (./a ./b)\n");
	    usage ();
	}
    }

    /* trim trailing / (needed for lnMode, nice otherwise) */
    {
	char *cp;
	for (cp = first; *cp; ++cp);	/* go to the end */
	if (cp > first) {
	    --cp;
	    if (*cp == '/') *cp = 0;	/* if its / then trim it */
	}
	for (cp = second; *cp; ++cp);	/* go to the end */
	if (cp > second) {
	    --cp;
	    if (*cp == '/') *cp = 0;	/* if its / then trim it */
	}
    }
    *matchPtr = NULL; *noMatchPtr = NULL; /* terminate match lists */

    if (!sawNewd && !noDef) {		/* use mkdir by default */
	newdCmd[i=0] = "mkdir";		/* default new dir command */
	newdCmd[++i] = secondFile;
	newdCmd[++i] = NULL;
    }
    if (!sawUpdate && !noDef) {		/* use copy by default */
	updateCmd[i=0] = fastCopy;	/* default update command */
	updateCmd[++i] = NULL;
    }
    if (!sawNew && !noNew) {	/* make new the same as update */
	char	**to;
	for (to = newCmd, cpp = updateCmd; (*to++ = *cpp++); );
    }
    if (verbose > 2) {
	printf ("Read: maptree %s %s ", first, second);
	for (matchPtr = matchExpr; *matchPtr; ++matchPtr) {
	    printf (" -match %s", *matchPtr);
	}
	for (matchPtr = noMatchExpr; *matchPtr; ++matchPtr) {
	    printf (" -nomatch %s", *matchPtr);
	}
	if (*updateCmd) {
	    printf (" -update ");
	    for (cpp = updateCmd; *cpp; ++cpp) {
		printf ("%s ", *cpp); fflush (stdout);
	    }
	}
	if (*newCmd) {
	    printf (" -new ");
	    for (cpp = newCmd; *cpp; ++cpp) {
		printf ("%s ", *cpp);
	    }
	}
	if (*newdCmd) {
	    printf (" -newd ");
	    for (cpp = newdCmd; *cpp; ++cpp) {
		printf ("%s ", *cpp);
	    }
	}
	if (*allCmd) {
	    printf (" -all ");
	    for (cpp = allCmd; *cpp; ++cpp) {
		printf ("%s ", *cpp);
	    }
	}
	printf ("\n");
    }
    MapTree (first, second, maxDepth);

    exit (0);
}

int
ParseCmd ( int	argc,
	   char	**argv,
	   char	**argList)		/* command to fill in */
{
    int		cnt, sawFirst=0;
    char	**thisArg = argList;

    for (cnt = 0; cnt < argc; ++cnt, ++argv) {
	if (**argv == ';') {	/* command ends at a semicolon */
	    *thisArg = NULL;
	    return (++cnt);
	} else if (!strcmp (*argv, "{}")) { /* handle {} {} */
	    if (!sawFirst) {
		*thisArg++ = firstFile;
		sawFirst++;
	    } else {
		*thisArg++ = secondFile;
	    }
	} else if (!strcmp (*argv, "{}1")) {
	    *thisArg++ = firstFile;
	} else if (!strcmp (*argv, "{}2")) {
	    *thisArg++ = secondFile;
	} else if (!strcmp (*argv, "{}3")) {
	    *thisArg++ = thirdFile;
	} else {
	    *thisArg++ = *argv;
	}
    }
    return (-1);		/* never found the end */
}

int
MapTree (char	*dir,
	 char	*dir2,
	 int	maxDepth)
{
    DIR		*dirList;
    struct dirent	*dirEnt;
    struct stat		statInfo, statInfo2;
    char	buff[MAXFILE], buff2[MAXFILE], buff3[MAXFILE], berror[MAXFILE];
    char	**matchPtr;
    int		ret, matched, chaseLink=0;
	
    if (stat (dir2, &statInfo) < 0) {	/* check if dir2 exists */
	sprintf (berror, "stat: %s", dir2);
	perror (berror);
	return (-1);
    }
    if (!(dirList = opendir (dir))) {
	sprintf (berror, "opendir: %s", dir);
	perror (berror);
	return (-1);
    }
    /* read through the directory list */
    for (; (dirEnt = readdir (dirList));) {
	if ((!strcmp (dirEnt->d_name, ".")) ||
	    (!strcmp (dirEnt->d_name, "..")) )
	  continue;
	
	sprintf (buff, "%s/%s", dir, dirEnt->d_name);
	/* Check match and nomatch strings */
	for (matched=0, matchPtr = matchExpr; *matchPtr; ++matchPtr) {
	    if (SubMatch (*matchPtr, dirEnt->d_name)) {
		matched=1;
		break;
	    }
	}
	if (*matchExpr && !matched) {
	    if (verbose > 1)
		printf ("Skipping un-matched entry: %s\n", dirEnt->d_name);
	    goto skip_this_one;
	}
	for (matchPtr = noMatchExpr; *matchPtr; ++matchPtr) {
	    if (SubMatch (*matchPtr, dirEnt->d_name)) {
		if (verbose > 1)
		    printf ("Skipping No-Match entry: %s\n", dirEnt->d_name);
		goto skip_this_one;
	    }
	}

	if (lstat (buff, &statInfo) < 0) { /* get first file type */
	    sprintf (berror, "lstat: %s", buff);
	    perror (berror);
	    continue;
	}
	if (S_ISDIR (statInfo.st_mode)) { /* do sub directories later  */
	    continue;
	}

#ifdef S_ISLNK
	if (S_ISLNK (statInfo.st_mode)) {
#else
	if (S_IFLNK == statInfo.st_mode) {
#endif
	    struct stat		statInfo3;
	    if (noSym) {		/* skip all symbolic links */
		if (verbose) printf ("Skipping symbolic link %s\n", buff);
		continue;
	    }
	    if (stat (buff, &statInfo3) < 0) { /* get real file file type */
		sprintf (berror, "stat3: %s", buff);
		perror (berror);
		continue;
	    }
	    if (S_ISDIR (statInfo3.st_mode)) {
	    /* treat links to dirs as a link iff following links */
		if (symDir) {
		    continue;
		}
		chaseLink = 0;
	    } else if (symFile) {	/* file, use dest file statinfo */
		bcopy (&statInfo3, &statInfo, sizeof (statInfo));
		chaseLink = 1;
	    }
	}
					/* find second file */
	buff2[0] = 0; buff3[0] = 0;
	sprintf (buff2, "%s/%s", dir2, dirEnt->d_name);
	ret = ((symFile && chaseLink)
	       ? stat (buff2, &statInfo2)
	       : lstat (buff2, &statInfo2));
	if (ret >= 0) strcpy (buff3, buff2); /* note matching name */
	if ((ret < 0) && fileExt) { /* try with extension */
	    sprintf (buff3, "%s/%s%s", dir2, dirEnt->d_name, fileExt);
	    ret = ((symFile && chaseLink)
		   ? stat (buff3, &statInfo2)
		   : lstat (buff3, &statInfo2));
	}
	if ((ret < 0) && fileRext && *fileRext) { /* try reverse extension */
	    char *to, *from;
	    strcpy (buff3, buff2);
	    for (to=buff3; *to; ++to); /* go to the end */
	    for (from=fileRext; *from; ++from); /* go to the end */
	    for (--from, --to;
		 (from >= fileRext) && (to >= buff3) && (*to == *from);
		 --to, --from);
	    if ((from < fileRext) && (to >= buff3)) { /* matched */
		to[1] = 0;	/* truncate name to drop extension */
		ret = ((symFile && chaseLink)
		       ? stat (buff3, &statInfo2)
		       : lstat (buff3, &statInfo2));
	    }
	}
					/* now do something with them */
	if (ret < 0) {			/* couldn't stat, new file */
	    if (verbose > 1) printf ("New file %s from %s\n", buff2, buff);
	    if (((newCmd[0] == fastCopy)
		 ? ((lnMode == 0)
		    ? Copy (buff, buff2, 1, statInfo)
		    : Link (buff, buff2))
		 : RunIt (buff, buff2, newCmd, buff3)) != 0) {
		fprintf (stderr, "maptree: creation failed for %s\n", buff2);
	    }
	} else {
	    if (lnMode == 0) {		/* don't try to update links for now */
		/* check if we are comparing files to files and dirs to dirs */
		/* BUG? The 0xf000 could be OS dependent */
		if ((statInfo.st_mode & 0xf000)
		    != (statInfo2.st_mode & 0xf000)) {
		    fprintf (stderr,
			     "Warning! Differing types: %s(%x)  %s(%x)\n",
			     buff, statInfo.st_mode, buff2, statInfo2.st_mode);
		}
	    
		if (statInfo.st_mtime > statInfo2.st_mtime) {
		    if (verbose > 1) printf ("Update file %s (%s)from %s\n",
					     buff2, buff3, buff);
		    if (((newCmd[0] == fastCopy)
			 ? Copy (buff, buff2, 0, statInfo)
			 : RunIt (buff, buff2, updateCmd, buff3)) != 0) {
			fprintf (stderr,
				 "maptree: update failed for %s\n", buff2);
		    }
		} else if (verbose > 2) {
		    printf ("SAME: %s %s\n", buff, buff2);
		}
	    }
	}
	if (allCmd[0]) RunIt (buff, buff2, allCmd, buff3);
      skip_this_one: ;			/* so we can break inner loops */
    }

    if (maxDepth <= 0) return (0);	/* don't do more sub directories */

    rewinddir (dirList);
    /* read through directory list for directories */
    for (; (dirEnt = readdir (dirList));) {
	if ((!strcmp (dirEnt->d_name, ".")) ||
	    (!strcmp (dirEnt->d_name, "..")) )
	  continue;

	sprintf (buff, "%s/%s", dir, dirEnt->d_name);

					/* check match strings */
	for (matched=0, matchPtr = matchExpr; *matchPtr; ++matchPtr) {
	    if (SubMatch (*matchPtr, dirEnt->d_name)) {
		matched=1;
		break;
	    }
	}
	if (*matchExpr && !matched) {
	    if (verbose > 1)
		printf ("Skipping un-matched dir: %s\n", dirEnt->d_name);
	    goto skip_this_dir;
	}
	for (matchPtr = noMatchExpr; *matchPtr; ++matchPtr) {
	    if (SubMatch (*matchPtr, dirEnt->d_name)) {
		if (verbose > 1)
		    printf ("Skipping No-Match dir: %s\n", dirEnt->d_name);
		goto skip_this_dir;
	    }
	}
		
	if (((symDir)
	     ? stat (buff, &statInfo)
	     : lstat (buff, &statInfo)) < 0) { /* get file type */
	    sprintf (berror, "(l)stat: %s", buff);
	    perror (berror);
	    continue;
	}
	if (!(S_ISDIR (statInfo.st_mode))) { /* only sub directories */
	    continue;
	}

	sprintf (buff2, "%s/%s", dir2, dirEnt->d_name);
	ret = stat (buff2, &statInfo2);
	if (ret < 0) {			/* make the new directory */
	    if (lnMode <= 1) {
		if (verbose > 1) printf ("New dir %s from %s\n", buff2, buff);
		/* we should have built in mkdir that copies permissions */
		RunIt (buff, buff2, newdCmd, buff2);
		if (stat (buff2, &statInfo2) < 0) {
		    fprintf (stderr,
			     "maptree: directory creation failed for %s\n",
			     buff2);
		    continue;
		}
	    } else {			/* make a link to it instead */
		Link (buff, buff2);
		goto skip_this_dir;	/* don't decend into source dir */
	    }
	}
	MapTree (buff, buff2, maxDepth-1);
      skip_this_dir: ;			/* so we can break inner loops */
    }
    closedir (dirList);
    return (0);
}

int
RunIt (char *file1, char *file2, char **cline, char *file3)
{
    int		status, child, ret, printed=0;
    char	*cmd[MAXARG], **from, **to;

    if (!cline[0]) return (0);	/* nothing to run */
    /* setup cmd list, copy args or substitute file names */
    for (from = cline, to = cmd; *from; ++from, ++to) {
	if (*from == firstFile) {
	    *to = file1;
	    if (verbose || noExec) printf ("\t%s ", *to);
	} else if (*from == secondFile) {
	    *to = file2;
	    if (verbose || noExec) printf ("\t%s ", *to);
	} else if (*from == thirdFile) {
	    *to = file3;
	    if (verbose || noExec) printf ("\t%s ", *to);
	} else {
	    *to = *from;
	    if (verbose || noExec) printf ("%s ", *to);
	}
    }

    *to = NULL;
    if (verbose || noExec) {printf ("\n"); printed = 1;}

    if (noExec) return (0);
    /* fork process */
    child = fork ();
    if (child) {
	/* parent, wait until done and get status */
	while ((waitpid (child, &status, 0) == -1) && (errno == EINTR));

	if (!(status & 0xff)) {
	    ret = (char)((status >> 8) & 0xff);
	} else {
	    ret = 9999;
	}
    } else {			/* child, exec the new process */
	execvp (cmd[0], cmd);
	exit (-1);
    }
    if (ret != 0) {			/* show failure */
	if (!printed) {			/* print command */
	    for (from = cmd; *from; ++from) fprintf (stderr, "%s ", *from);
	} else fflush (stdout);		/* make sure verbose shown */
	if (ret == 9999) {
	    fprintf (stderr, "was killed.\n");
	} else {
	    fprintf (stderr, "returned %d.\n", ret);
	}
    }
    return (ret);
}

/* Recursive string matching supporting * and ? wildcards */
int
SubMatch ( char	*pattern,
	   char *string)
{
    char	*src, *dst;

    for (src = pattern, dst= string; *src || *dst; ) {
	if (*src == '*') {	/* match zero or more characters */
	    ++src;		/* next real character */
	    for (;;) {
		while (*dst && (*dst != *src)) ++dst;
		if (!*dst) break; /* end of string */
		if (SubMatch (src, dst)) break; /* match */
		else ++dst;	/* remainder did not fully match */
	    }
	} else if (*src == '?') { /* match any one character */
	    ++src; ++dst;
	} else {
	    if (*src == *dst) {
		++src;
		++dst;
	    } else {
		if (verbose > 4)
		  printf ("Mismatch of %s %s at <%s> <%s>\n",
			  pattern, string, src, dst);
		break;
	    }
	}
    }
    if (!*src && !*dst) return (1);	/* matched */
    return (0);				/* failed */
}

/* remove file to, copy file from to file to, match permission and owner */
int
Copy ( char *fromFile, char *toFile,
    int	newFile,			/* 1 if toFile does not exits */
    struct stat fromStat)
{
    static char	buff[65536];
    int		ret, from, to;
    struct utimbuf	modTime;

    if (!newFile && !noExec) {		/* file exists, remove it */
	if (unlink (toFile) < 0) {
	    sprintf (buff, "unlink: %s", toFile);
	    perror (buff);
	    return (-1);
	}
    }
					/* duplicate symbolic link */
#ifdef S_ISLNK
    if (S_ISLNK (fromStat.st_mode)) {
#else
    if (S_IFLNK == fromStat.st_mode) {
#endif
	if ((ret = readlink (fromFile, buff, sizeof (buff))) <= 0) {
	    sprintf (buff, "readlink: %s", fromFile);
	    perror (buff);
	    return (-4);
	}
	buff[ret] = 0;			/* null terminate buffer */
	if (verbose || noExec)
	    printf ("fln -s\t%s \t%s\n", buff, toFile);

	if (noExec) return (0);
	if (symlink (buff, toFile) < 0) {
	    sprintf (buff+512, "symlink %s %s", buff, toFile);
	    perror (buff+512);
	    return (-5);
	}
	return (0);
    }

    if (verbose || noExec) {
	printf ("fcp\t%s \t%s\n", fromFile, toFile);
    }
    if (noExec) return (0);
    if ((from = open (fromFile, O_RDONLY, 0)) < 0) { /* open from file first */
	sprintf (buff, "open: %s", fromFile);
	perror (buff);
	return (-2);
    }

    if ((to = open (toFile, O_WRONLY | O_CREAT, 0600)) < 0) { /* dest file */
	sprintf (buff, "open2: %s", toFile);
	perror (buff);
	close (from);
	return (-3);
    }
    
					/* copy the file */
    while ((ret = read (from, buff, sizeof (buff))) > 0) {
	if (write (to, buff, ret) < 0) {
	    perror ("write");
	    break;
	}
    }
    if (ret < 0) {
	sprintf (buff, "fcopy: %s", fromFile);
	perror (buff);
    }
    close (from);

    modTime.actime = fromStat.st_atime;
    modTime.modtime = fromStat.st_mtime;
    utime (toFile, &modTime);		/* clone times before change owner*/
    fchmod (to, fromStat.st_mode);	/* clone permissions */
    if (doOwners) {
	fchown (to, fromStat.st_uid, fromStat.st_gid); /* clone owner, group */
	if (fromStat.st_mode & (S_ISUID | S_ISGID)) {
	    fchmod (to, fromStat.st_mode); /* get suid/sgid bits back */
	}
    }
    utime (toFile, &modTime);		/* try to clone times (again) */
    
    close (to);
    return (0);
}

/* link file from to file to */
int
Link ( char *fromFile, char *toFile)
{
    static char	cwd[2*PATH_MAX+20];
    static char	relPath[PATH_MAX+1];
    char	*destDir, *lastSlash;
    char	*sameDir, *lastSameSlash;
    int		retVal = 0, slashCnt = 0;

    *cwd = 0;
    lastSameSlash = lastSlash = 0;
					/* figure out how much is the same */
    for (destDir = toFile, sameDir = fromFile;
	 *destDir && (*destDir == *sameDir);
	 ++destDir, ++sameDir)
	if (*sameDir == '/') lastSameSlash = sameDir;

					/* find out how much more there is */
    for (;*destDir; ++destDir)
	if (*destDir == '/') {
	    lastSlash = destDir;
	    ++slashCnt;
	}
    if (!slashCnt || !lastSlash) {
	printf ("Couldn't figure out ln -s %s %s.\n", fromFile, toFile);
	return -2;
    }
    getcwd (cwd, sizeof (cwd));
    *lastSlash = 0;
    if (chdir (toFile) < 0)  {		/* cd there */
	sprintf (cwd, "cd %s", toFile);
	perror (cwd);
	return -1;
    }
    if (*fromFile != '/') {		/* relative link, fix it up */
	int i;
	char *cp = relPath;
	if (!lastSameSlash) {
	    printf ("Couldn't figure out ln -s %s %s.\n", fromFile, toFile);
	    retVal = -3;
	    goto cleanReturn;
	}
	/*printf ("need to insert %d ../ to make it work. %s %s\n",
		slashCnt, fromFile, toFile);*/
	fflush (stdout);
	for (i = slashCnt; i-- > 0;) {
	    *cp++ = '.'; *cp++ = '.'; *cp++ = '/';
	}
	strcpy (cp, lastSameSlash+1);
    } else {
	strcpy (relPath, fromFile);
    }
    if (verbose || noExec) {
	printf ("fcd\t%s \t;fln -s %s %s\n",
		toFile, relPath, lastSlash+1);
    }
    *lastSlash++ = '/';		/* clean up and advance to file name */
    if (noExec) goto cleanReturn;
    if (symlink (relPath, lastSlash) < 0) {
	sprintf (cwd, "symlink %s %s", fromFile, toFile);
	perror (cwd);
	retVal = -5;
    }
    
cleanReturn:
    if (*cwd) chdir (cwd);
    return retVal;
}
