/*===========================================================================*\
|| Copyright 1989-1991 by Pittsburgh Powercomputing Corporation (PPc) 
|| Pittsburgh Pennsylvania as an unpublished work. All Rights Reserved.
||
|| This is unpublished proprietary trade secret source code of Pittsburgh
|| Powercomputing. The copyright above does not evidence any actual or intended
|| publication of this source code.  This source code is supplied under a
|| license.  It may be used, disclosed, and/or copied only as permitted under
|| such license agreement.  Use, copying, and/or disclosure of this source
|| code is strictly prohibited unless otherwise provided in the license
|| agreement.  Any copy must contain the copyright above and this notice must
|| be preserved in all copies of this source code.
\*===========================================================================*/
/*
 
 $Header:$
 $Id:$
 $Log:$

*/
/* Evaluate an expression and print the results */
#include <stdio.h>
#define skip_wht(s) do {++(*(s)); \
    while ((**(s) == ' ') || (**(s) == '\t')) ++(*(s)); } while (0)

#ifdef UNUM
#define NTYPE unsigned long
#else
#define NTYPE long
#endif

NTYPE evaluate(), colnum();
int parlevel, errors;

NTYPE evaluate(str)
  char **str;
{
    NTYPE tvalue,svalue;
    char op, c;
    
    op = '+';
    svalue = 0;
    while (c = **str) {
	tvalue = 0;
	
	if (c == '(') {		/* get a number */
	    skip_wht (str);
	    ++parlevel;
	    tvalue = evaluate(str);
	}
	else if (c == '%' || (c >= '0' && c <= '9')) /* number */
	  tvalue = colnum(str);
	
	else if (c == '\'') {	/* quoted character */
	    ++(*str);
	    tvalue = **str & 0xff;
	    skip_wht(str);
	}
	
	switch (op) {		/* operate on number */
	  case '+':
	    svalue += tvalue;
	    break;
	  case '-':
	    svalue -= tvalue;
	    break;
	  case '/':
	    svalue = (unsigned) svalue/tvalue;
	    break;
	  case '*':
	    svalue *= tvalue;
	    break;
	  case '%':
	    svalue = (unsigned) svalue % tvalue;
	    break;
	  case '^':
	    svalue ^= tvalue;
	    break;
	  case '~':
	    svalue = ~tvalue;
	    break;
	  case '&':
	    svalue &= tvalue;
	    break;
	  case '|':
	    svalue |= tvalue;
	    break;
	  case '<':
	    svalue = (svalue < tvalue) ? 1 : 0;
	    break;
	  case '>':
	    svalue = (svalue > tvalue) ? 1 : 0;
	    break;
	  default:
	    ++errors;
	}
	if (! (**str)) return (svalue);	/* hit end of expression */
	if ((op = **str) == ')') {
	    --parlevel;
	    skip_wht(str);
	    return(svalue);
	}
	skip_wht(str);
    }
    return(svalue);
}

/*
  ||	collect number operand
  */

NTYPE colnum(str)
  char	**str;
{
    NTYPE mul, nval;
    char c;
    
    nval = 0;
    if ((c = **str) == '0') {
	if ((*(*str +1)) == 'x') {
	    mul = 16;		/* hex 0x... */
	    skip_wht(str);
	} else mul = 8; }	/* octal 0... */
    else if (c >= '1' && c <= '9') { mul = 10; nval = c - '0';} /* deciml */
    else if (c == '%') mul = 2; /* binary %... */
    else { ++errors; return (0);};
    
    while ((c = (*(++(*str)) - '0')) >= 0) {
	if (c > 9) {
	    c -= ('A' - '9' - 1);
	    if (c > 15) c -= ('a' - 'A');
	    if (c < 10) break;
	}
	if (c >= mul) break;
	nval = (nval * mul) + (c & 0xff);
    }
    while ((**(str) == ' ') || (**(str) == '\t')) ++(*(str));
    return(nval);
}

main (argc, argv)
  int argc;
  char **argv;
{
    char *expr;
    NTYPE value, low, high, lowB=0, highB=0;
    int	numMode=0;

    for (--argc, ++argv; argc > 0; --argc, ++argv) {
	expr = *argv;
	parlevel = errors = 0;
	if (*expr == '-') {
	    switch (expr[1]) {
	      case 'h': numMode = 1; continue;
	      case 'd': numMode = 2; continue;
	      case 'o': numMode = 3; continue;
	      case 'a': numMode = 4; continue;
	      case 'H': numMode = 5; continue; /* hexidecimal, no 0x */
	      case 'O': numMode = 6; continue; /* ocatal, no 0 */
	      case 'l':		/* lower bounds check */
		if (!--argc) exit (2);
		expr = *(++argv);
		low = evaluate (&expr);
		lowB = 1;
		if (parlevel || errors) { exit (2); }
		continue;
	      case 'u':		/* upper bounds check */
		if (!--argc) exit (2);
		expr = *(++argv);
		high = evaluate (&expr);
		highB = 1;
		if (parlevel || errors) { exit (2); }
		continue;
	    }
	}
	value = evaluate (&expr);
				/* correct evalutation */
	if (errors || parlevel) {
	    if (!numMode) {
		printf ("Expression: %s.  %d errors.  %d paren mismatchs.\n",
			*argv, errors, parlevel);
		continue;
	    } else exit (1);
	}
				/* bounds check */
	if (lowB && (value < low)) exit (1);
	if (highB && (value > high)) exit (1);

	switch (numMode) {
	  case 0:
	    printf ("The value is:  %d \t%#x \t%#o\n", value, value, value);
	    break;
	  case 1: printf ("0x%x\n", value); break;
	  case 2: printf ("%d\n", value); break;
	  case 3: printf ("0%o\n", value); break;
	  case 4:
	    if (value > ' ') printf ("%c", value);
	    else printf ("^%c\n", value + '@');
	    break;
	  case 5: printf ("%x\n", value); break;
	  case 6: printf ("%o\n", value); break;
	}
    }
    exit (0);
}
