/* script to unwrap C function calls to a be on a single line
 Concatenates lines whenever there are more (s than )s
 Dan Christian 2-6-96 */

#include <stdio.h>

int debug = 0;

main (int argc, char **argv)
{
    char	oneLine[512];		/* incoming line */
    char 	allLine[8192];		/* our concatenated line */
    char	*from, *to;
    char	*matchStart, *match;
    int		openParen;
    int		continues;
    FILE	*inFile;

    allLine[0] = 0;			/* clear buffer */
    continues = openParen = 0;
    to = allLine;
    match = matchStart = 0;
    --argc, ++argv;			/* skip command */
    if (argc <= 0) {
	inFile = stdin;
    } else {
	inFile = fopen (*argv, "r");
	if (inFile == 0) exit (1);
    }
    while (fgets (oneLine, sizeof (oneLine), inFile)) {	/* while not eof */
	for (from = oneLine; *from; ++from) {
	    if (*from == '\\') {	/* handle quoting */
		*to++ = *from++;	/* the \ */
		if ((*from == '\n') || (*from == '\r')) continue; /* \nl */
		*to++ = *from;		/* quoted character */
		continue;
	    }
	    if (match) {
		if (*from == *match) {	/* string match */
		    ++match;
		    if (!*match) {	/* complete match */
			if (debug) printf ("END %s ", matchStart);
			match = matchStart = 0;
		    }
		} else match = matchStart; /* reset search */
		if ((*from == '\n') || (*from == '\r')) continue;
		*to++ = *from;
		if (to > (allLine + sizeof (allLine) - 2)) {
		    fprintf (stderr, "unformat: overrun! %d <%s>\n",
			     openParen, matchStart);
		    exit (1);
		}
		continue;
	    }

	    /* check if we are in initial white space */
	    if (continues && ((*from == ' ') || (*from == '\t'))) {
		continue;
	    }
	    if (continues) {		/* insert a single space */
		*to++ = ' ';
		continues = 0;
	    }

	    /* check for comments, strings, and count parens */
	    switch (*from) {
	    case '/':			/* comment */
		if (from[1] == '*') {	/* BUG accepts slash-star-slash */
		    match = matchStart = "*/";
		}
		break;
		  
	    case '"':		/* string */
		if (debug) printf ("Start \" ", matchStart);
		match = matchStart = "\"";
		break;
		  
	    case '\'':		/* char */
		match = matchStart = "\'";
		break;

	    case '(':
		++openParen;
		break;

	    case ')':
		--openParen;
		break;

	    case '\r':			/* skip these */
	    case '\n':			/* skip these (re-insert on print) */
		continue;

	    }
	    *to++ = *from;		/* put into the big buffer */
	    if (to > (allLine + sizeof (allLine) - 2)) {
		fprintf (stderr, "unformat: overrun! %d\n", openParen);
		exit (1);
	    }
	}

	/* done checking the line, figure out what to do with it */
	if (openParen <= 0) {		/* nothing open */
	    *to = 0;			/* terminate string */
	    puts (allLine);		/* add newline back in */
	    fflush (stdout);
	    to = allLine;		/* clear buffer */
	} else {			/* continue the line */
	    continues = 1;
	}
    }
    return 0;
}
